package pps.core.space.service;

import cn.hutool.core.collection.CollUtil;
import com.baomidou.mybatisplus.core.conditions.query.LambdaQueryWrapper;
import org.apache.commons.lang3.StringUtils;
import pps.cloud.base.service.IBasePowerLineCloudService;
import pps.cloud.base.service.IBaseWellheadCloudService;
import pps.cloud.base.service.data.base_power_line_plant.DynamicQueryBasePowerLinePlantInput;
import pps.cloud.base.service.data.base_power_line_plant.DynamicQueryBasePowerLinePlantOutput;
import pps.cloud.base.service.data.base_wellhead.DynamicQueryBaseWellheadInput;
import pps.cloud.base.service.data.base_wellhead.DynamicQueryBaseWellheadOutput;
import pps.cloud.prediction.service.IPlantPredictedPowerCloudService;
import pps.cloud.prediction.service.data.plant_predicted_power_data.DynamicQueryPlantPredictedPowerInput;
import pps.cloud.prediction.service.data.plant_predicted_power_data.DynamicQueryPlantPredictedPowerOutput;
import pps.core.common.constant.BusinessConstant;
import pps.core.common.entity.BaseModel;
import pps.core.space.entity.SpaceInstitutionDetailEnt;
import pps.core.space.entity.SpaceInstitutionDurationEnt;
import pps.core.space.entity.SpaceInstitutionWellheadView;
import pps.core.space.enums.BusinessError;
import pps.core.space.mapper.SpaceInstitutionDetailMapper;
import pps.core.space.mapper.SpaceInstitutionDurationMapper;
import pps.core.space.mapper.SpaceInstitutionWellheadViewMapper;
import pps.core.space.service.data.SpaceOptimizeWellheadAndPlant;
import xstartup.base.XContext;
import xstartup.base.exception.XServiceException;
import xstartup.data.XListResult;

import java.math.BigDecimal;
import java.util.*;
import java.util.stream.Collectors;

/**
 * 间开优化公用方法
 *
 * @author ZWT
 * @date 2023/09/18 16:02
 */
public class SpaceOptimizeBaseService {

    /**
     * 获得生效中的基础间开
     *
     * @param context 上下文
     * @param sameDay 同一天
     * @return {@link List}<{@link SpaceInstitutionDetailEnt}>
     */
    public List<SpaceInstitutionDetailEnt> getEffectiveSpaceInstitution(XContext context, String sameDay) {
        //取生效中的基础间开
        SpaceInstitutionDetailMapper detailMapper = context.getBean(SpaceInstitutionDetailMapper.class);
        List<SpaceInstitutionDetailEnt> detailEntList = detailMapper.selectList(new LambdaQueryWrapper<SpaceInstitutionDetailEnt>()
                .eq(BaseModel::getIsDeleted, BusinessConstant.ONE)
                .eq(SpaceInstitutionDetailEnt::getIsCurrentBasic, BusinessConstant.ZERO)
                .le(SpaceInstitutionDetailEnt::getInstitutionStartDate, sameDay)
                .ge(SpaceInstitutionDetailEnt::getInstitutionEndDate, sameDay)
        );
        if (CollUtil.isEmpty(detailEntList)) {
            //未发现可优化基础间开制度
            throw new XServiceException(BusinessError.DidNotFindSpace);
        }
        return detailEntList;
    }

    /**
     * 获取所有生效中制度关联的井口及光伏电站
     *
     * @param context       上下文
     * @param detailEntList 详细ent列表
     * @return {@link SpaceOptimizeWellheadAndPlant}
     */
    public SpaceOptimizeWellheadAndPlant getWellheadAndPlant(XContext context, List<SpaceInstitutionDetailEnt> detailEntList) {
        Set<String> lineIds = new HashSet<>(32);
        Set<String> institutionIds = new HashSet<>(32);
        for (SpaceInstitutionDetailEnt spaceInstitutionDetailEnt : detailEntList) {
            lineIds.add(spaceInstitutionDetailEnt.getLineId());
            institutionIds.add(spaceInstitutionDetailEnt.getId());
        }
        //取生效中间开的所有井口
        SpaceInstitutionWellheadViewMapper wellheadMapper = context.getBean(SpaceInstitutionWellheadViewMapper.class);
        List<SpaceInstitutionWellheadView> spaceWellheadList = wellheadMapper.selectListByInstitutionIds(institutionIds);
        if (CollUtil.isEmpty(spaceWellheadList)) {
            //未发现可优化井口
            throw new XServiceException(BusinessError.DidNotFindWellhead);
        }
        this.setServiceRatingForSpaceWellheadList(context, spaceWellheadList);
        //取线路关联的所有光伏电站ID
        List<DynamicQueryBasePowerLinePlantOutput> plantList = this.getPowerLinePlantListByLineIds(context, lineIds);
        if (CollUtil.isEmpty(plantList)) {
            //未发现可用光伏电站
            throw new XServiceException(BusinessError.DidNotFindPlant);
        }
        return SpaceOptimizeWellheadAndPlant.builder()
                .spaceWellheadList(spaceWellheadList)
                .plantIdsByLineIdMap(
                        plantList.stream()
                                .collect(
                                        Collectors.groupingBy(DynamicQueryBasePowerLinePlantOutput::getLineId,
                                                Collectors.mapping(DynamicQueryBasePowerLinePlantOutput::getPlantId, Collectors.toList()))
                                )
                )
                .build();
    }

    /**
     * 条件获取获取光伏预测各时段平均值列表
     *
     * @param context 上下文
     * @param input   输入
     * @return {@link List}<{@link DynamicQueryPlantPredictedPowerOutput}>
     */
    public List<DynamicQueryPlantPredictedPowerOutput> getAveragePowerGenerationListByPlantIds(XContext context, DynamicQueryPlantPredictedPowerInput input) {
        IPlantPredictedPowerCloudService cloudService = context.getBean(IPlantPredictedPowerCloudService.class);
        XListResult<DynamicQueryPlantPredictedPowerOutput> result = cloudService.queryAveragePowerGenerationListByParam(context, input);
        result.throwIfFail();
        return result.getResult();
    }

    /**
     * 获得最大功率
     *
     * @param avgPowerList 平均功率列表
     * @return {@link BigDecimal}
     */
    public BigDecimal getPowerMax(List<DynamicQueryPlantPredictedPowerOutput> avgPowerList) {
        return avgPowerList.stream()
                .map(DynamicQueryPlantPredictedPowerOutput::getPower)
                .max(BigDecimal::compareTo)
                .orElse(BigDecimal.ZERO);
    }

    /**
     * 获取井口总功率
     *
     * @param spaceWellheadList 空间井口清单
     * @param detailId          详细信息id
     * @return {@link BigDecimal}
     */
    public BigDecimal getWellheadTotalPower(List<SpaceInstitutionWellheadView> spaceWellheadList, String detailId) {
        return spaceWellheadList.stream()
                .filter(w -> StringUtils.equals(detailId, w.getInstitutionId()))
                .map(SpaceInstitutionWellheadView::getServiceRating)
                .reduce(BigDecimal.ZERO, BigDecimal::add);
    }

    /**
     * 获取井口视图列表
     *
     * @param spaceWellheadList 空间井口清单
     * @param detailId          详细信息id
     * @return {@link List}<{@link SpaceInstitutionWellheadView}>
     */
    public List<SpaceInstitutionWellheadView> getWellheadViewList(List<SpaceInstitutionWellheadView> spaceWellheadList, String detailId) {
        return spaceWellheadList.stream()
                .filter(w -> StringUtils.equals(detailId, w.getInstitutionId()) &&
                        StringUtils.equals("INTERVAL", w.getRunTypeKey()) &&
                        StringUtils.equals("0", w.getIntervalTypeKey()))
                .sorted(Comparator.comparing(SpaceInstitutionWellheadView::getStartSeq))
                .collect(Collectors.toList());
    }

    /**
     * 获取井口间开时间段
     *
     * @param durationMapper 持续时间映射器
     * @param detailId       详细信息id
     * @return {@link Map}<{@link String}, {@link List}<{@link SpaceInstitutionDurationEnt}>>
     */
    public Map<String, List<SpaceInstitutionDurationEnt>> getDurationMap(SpaceInstitutionDurationMapper durationMapper, String detailId) {
        //通过间开ID和井口ID查所有井口时段配置
        List<SpaceInstitutionDurationEnt> durationList = durationMapper.selectList(new LambdaQueryWrapper<SpaceInstitutionDurationEnt>()
                .eq(BaseModel::getIsDeleted, BusinessConstant.ONE)
                .eq(SpaceInstitutionDurationEnt::getInstitutionId, detailId)
                .orderByAsc(SpaceInstitutionDurationEnt::getOpenWellTime)
        );
        if (CollUtil.isEmpty(durationList)) {
            //没有设置时间段,无法优化
            return null;
        }
        return durationList.stream()
                .collect(
                        Collectors.groupingBy(SpaceInstitutionDurationEnt::getWellheadId)
                );
    }

    /**
     * 获取井口Map通过最大功率区分
     *
     * @param wellheadViewList 井口视图列表
     * @param powerMax         最大功率
     * @return {@link Map}<{@link Boolean}, {@link List}<{@link SpaceInstitutionWellheadView}>>
     */
    public Map<Boolean, List<SpaceInstitutionWellheadView>> getWellheadViewMapByPower(List<SpaceInstitutionWellheadView> wellheadViewList, BigDecimal powerMax) {
        return wellheadViewList.stream()
                .collect(
                        Collectors.partitioningBy(w -> powerMax.compareTo(w.getServiceRating()) >= BusinessConstant.ZERO)
                );
    }

    /*-----------------------------------private-----------------------------------*/

    /**
     * 按线路ID集合获取电站列表
     *
     * @param context 上下文
     * @param lineIds 线路ID
     * @return {@link List}<{@link DynamicQueryBasePowerLinePlantOutput}>
     */
    private List<DynamicQueryBasePowerLinePlantOutput> getPowerLinePlantListByLineIds(XContext context, Set<String> lineIds) {
        IBasePowerLineCloudService cloudService = context.getBean(IBasePowerLineCloudService.class);
        XListResult<DynamicQueryBasePowerLinePlantOutput> result = cloudService.queryPowerLinePlantListByParam(context,
                DynamicQueryBasePowerLinePlantInput.builder()
                        .lineIds(lineIds)
                        .build()
        );
        result.throwIfFail();
        return result.getResult();
    }

    /**
     * 为间开井口列表设置发电功率
     *
     * @param context           上下文
     * @param spaceWellheadList 空间井口清单
     */
    private void setServiceRatingForSpaceWellheadList(XContext context, List<SpaceInstitutionWellheadView> spaceWellheadList) {
        IBaseWellheadCloudService cloudService = context.getBean(IBaseWellheadCloudService.class);
        XListResult<DynamicQueryBaseWellheadOutput> result = cloudService.queryBaseWellheadListByParam(context,
                DynamicQueryBaseWellheadInput.builder()
                        .wellheadIds(spaceWellheadList.stream()
                                .map(SpaceInstitutionWellheadView::getWellheadId)
                                .collect(Collectors.toList())
                        )
                        .build());
        result.throwIfFail();
        List<DynamicQueryBaseWellheadOutput> wellheadOutputList = result.getResult();
        if (CollUtil.isEmpty(wellheadOutputList)) {
            return;
        }
        spaceWellheadList.forEach(s -> {
            s.setServiceRating(
                    wellheadOutputList.stream()
                            .filter(w -> StringUtils.equals(s.getWellheadId(), w.getId()))
                            .findAny()
                            .map(DynamicQueryBaseWellheadOutput::getServiceRating)
                            .orElse(BigDecimal.ZERO)
            );
        });
    }
}
